#include <Wire.h>
#include "SSD1306Ascii.h"
#include "SSD1306AsciiWire.h"
#include <EEPROM.h>
#include <SPI.h>
#include <SD.h>

// U W A G A !
// Przed wgraniem kodu do licznika prosze sprawdzic informacje na poczatku funkcji setup().
// A T T E N T I O N !
// Before uploading the code into the counter please check the information at the begining of setup() function.

#define VERSION F("19.10.07")
#define LED_PIN 3
#define COINCIDENCE_PIN 6
#define LED_BRIGHTNESS 250 // [0...255]
#define TIMER_INTERVAL 1000 // ms
#define SLOPE_THRESHOLD 25
#define DETECTOR_NAME_SIZE 20
#define SD_CARD_DELAY 2 // ms

#define SDPIN 10
Sd2Card card;
SdVolume volume;
SdFile root;

const float cal[] = { 7.951e-14, -1.529e-10, 1.092e-7, -3.527e-5, 5.830e-3, -3.545e-1, 4.054e1 };

SSD1306AsciiWire oled;
File my_file;
char detector_name[ DETECTOR_NAME_SIZE ];
unsigned long total_counts = 0;
unsigned long start_time = 0;
unsigned long total_deadtime = 0;
unsigned long start_deadtime = 0;
unsigned long stop_deadtime = 0;
unsigned long current_time = 0;
unsigned long screen_time = TIMER_INTERVAL;
byte master = 0xff;
byte keep_pulse = 0x00;
byte use_card = 0xff;
unsigned int x0, x1, amplitude;
float count_average = 0.;
float count_std = 0.;
float float_time;
float voltage;
int seconds, minutes, hours;
int decimals;
char min_char[4], sec_char[4], average_char[8], stddev_char[8];
char filename[] = "File_000.txt";
char mode;
char message;
unsigned int tmp_u = 0;

void setup()
{
// Prosze usunac ("zakomentowac") ponizsza instrukcje, jesli licznik jest w wersji pierwszej (zielony lub pomarańczowy, bez dostępu do karty SD).
// Prosze przywrocic ("odkomentowac") ja, gdy licznik jest w wersji drugiej (niebieski, z kartą SD dostępną na zewnątrz).
// Comment the following line if the counter is the first version (green or orange, without the SD card slot on the backpanel).
// Uncomment it if the counter is the second version (blue, with the SD card slot on the backpanel).
  analogReference( EXTERNAL );

  ADCSRA &= ~(bit (ADPS0) | bit (ADPS1) | bit (ADPS2));     // clear prescaler bits
  ADCSRA |= bit (ADPS1);                                    // set prescaler to 4

  Serial.begin( 9600 );
  while( !Serial ) { ; }

  Wire.begin();
  Wire.setClock( 400000L );
  oled.begin( &Adafruit128x32, 0x3C );
  oled.setFont( System5x7 );
  oled.displayRemap( true );
  oled.clear();

  pinMode( LED_PIN, OUTPUT );
  pinMode( COINCIDENCE_PIN, INPUT );
  if( digitalRead( COINCIDENCE_PIN ) == HIGH ) 
  {
      master = 0x00;
      digitalWrite( LED_PIN, HIGH );
  }
  else 
  {
      pinMode( COINCIDENCE_PIN, OUTPUT );
      digitalWrite( COINCIDENCE_PIN, HIGH );
  }
  delay( 2000 );
  if( master )
    digitalWrite( COINCIDENCE_PIN, LOW );
  else
    digitalWrite( LED_PIN, LOW );

  if ( !SD.begin( SDPIN ) )
  {
    use_card = 0x00;
    Serial.println( F("SD card error!") );
    filename[0] = 'S';
    filename[1] = 'D';
    filename[2] = 'c';
    filename[3] = 'a';
    filename[4] = 'r';
    filename[5] = 'd';
    filename[6] = ' ';
    filename[7] = 'e';
    filename[8] = 'r';
    filename[9] = 'r';
    filename[10] = 'o';
    filename[11] = 'r';
  }
  delay( SD_CARD_DELAY );

  get_detector_name( detector_name );

  mode = 'N';
}

void loop()
{
  if( Serial.available() )
    message = Serial.read();
  else
    message = ' ';
  if( message == 'N' )
    mode = 'N';
  else
  if( message == 'R' )
    mode = 'R';
  else
  if( message == 'D' )
    mode = 'D';

  if( mode == 'N' )
  {
    if( my_file )
      my_file.close();
    if( use_card )
    {
      while( SD.exists( filename ) )
      {
        delay( SD_CARD_DELAY );
        next_filename();
      }
      my_file = SD.open( filename, FILE_WRITE );
      my_file.println( F("#########################################################") );
      my_file.println( F("### CosmicWatch: The Desktop Muon Detector") );
      my_file.println( F("### Arduino code: l.adamowski@ncbj.gov.pl") );
      my_file.println( F("### Event Ardn_time[ms] ADC[0-1023] SiPM[mV] Deadtime[ms]") );
      my_file.println( F("#########################################################") );
      my_file.println( "DetectorID: " + (String)detector_name );
      if( master )
        my_file.println( F("DetectorMode: Master") );
      else
        my_file.println( F("DetectorMode: Slave") );
    }
    Serial.println( F("#########################################################") );
    Serial.println( F("### CosmicWatch: The Desktop Muon Detector") );
    Serial.println( F("### Arduino code: l.adamowski@ncbj.gov.pl") );
    Serial.println( F("### Event Ardn_time[ms] ADC[0-1023] SiPM[mV] Deadtime[ms]") );
    Serial.println( F("#########################################################") );
    Serial.println( "DetectorID: " + (String) detector_name );
    if( master )
      Serial.println( F("DetectorMode: Master") );
    else
      Serial.println( F("DetectorMode: Slave") );
    mode = 'W';
    oled.clear();
    oled.println( F("CosmicWatch") );
    oled.println( detector_name );
    oled.print( F("ver.") );
    oled.println( VERSION );
    oled.println( filename );
    delay( 3000 );
    oled.setCursor( 0, 0 );
    if( master )
      oled.print( F("Master") );
    else
      oled.print( F(" Slave") );
    oled.println( F(" counts ") );
    oled.print( F("Uptime ") );
    x1 = analogRead( A0 );
    total_counts = 0;
    total_deadtime = 0;
    start_time = millis();
  }
  else
  if( mode == 'R' )
  {
    if( use_card )
    {
      if( my_file )
        my_file.close();
      oled.clear();
      oled.print( F("Reading ") );
      filename[5] = '0';
      filename[6] = '0';
      filename[7] = '0';
      tmp_u = 0;
      do {
        oled.setCursor( tmp_u/10, 1 );
        oled.print( F("_") );
        tmp_u++;
        oled.setCursor( 48, 0 );
        if( SD.exists( filename ) )
        {
          oled.print( filename );
          Serial.print( F("Reading: ") );
          Serial.println( filename );
          my_file = SD.open( filename );
          while( my_file.available() )
            Serial.write( my_file.read() );
          my_file.close();
          Serial.println( F("EOF") );
        }
        else
          oled.print( F("            ") );
      } while( !next_filename() );
      Serial.println( F("Done!") );
      mode = 'N';
    }
    else
      Serial.println( F("SD card unavailable!") );
  }
  else
  if( mode == 'D' )
  {
    if( use_card )
    {
      if( my_file )
        my_file.close();
      oled.clear();
      oled.print( F("Deleting ") );
      filename[5] = '0';
      filename[6] = '0';
      filename[7] = '0';
      tmp_u = 0;
      do {
        oled.setCursor( tmp_u/10, 1 );
        oled.print( F("_") );
        tmp_u++;
        oled.setCursor( 54, 0 );
        if( SD.exists( filename ) )
        {
          oled.print( filename );
          Serial.print( F("Deleting: ") );
          Serial.println( filename );
          SD.remove( filename );
        }
        else
          oled.print( F("            ") );
      } while( !next_filename() );
      Serial.println( F("Done!") );
      mode = 'N';
    }
    else
      Serial.println( F("SD card unavailable!") );
  }  
  else
  if( mode == 'W' )
  {
    x0 = x1;
    x1 = analogRead( A0 );
    if( x1 > x0 + SLOPE_THRESHOLD )
    {
      start_deadtime = millis();
      if( master )
      {
        digitalWrite( COINCIDENCE_PIN, HIGH );
        keep_pulse = 0xff;
      }
      amplitude = analogRead( A0 );
      if( ~master )
        if( digitalRead( COINCIDENCE_PIN ) == HIGH )
          keep_pulse = 0xff;
        else
          keep_pulse = 0x00;
      if( master )
        digitalWrite( COINCIDENCE_PIN, LOW );
      if( keep_pulse )
      {
        analogWrite( LED_PIN, LED_BRIGHTNESS );
        total_counts++;
        voltage = get_sipm_voltage( amplitude );
        Serial.println( (String) total_counts + " " + (start_deadtime - start_time) + " " + amplitude + " " + voltage + " " + total_deadtime );
        if( use_card )
        {
          my_file.println( (String) total_counts + " " + (start_deadtime - start_time) + " " + amplitude + " " + voltage + " " + total_deadtime );
          my_file.flush();
        }
        digitalWrite( LED_PIN, LOW );
      }
      stop_deadtime = millis();
      if( stop_deadtime < start_deadtime )
      {
        stop_deadtime += 2147483648;
        start_deadtime -= 2147483648;
      }
      total_deadtime += stop_deadtime - start_deadtime;
    }
    start_deadtime = millis();
    if( start_deadtime > screen_time )
    {
      oled.setCursor( 84, 0 );
      oled.print( total_counts );
      current_time = start_deadtime - start_time;
      seconds = current_time / 1000;
      minutes = seconds / 60;  
      hours = minutes / 60;
      minutes %= 60;
      seconds %= 60;
      sprintf( min_char, "%02d", minutes );
      sprintf( sec_char, "%02d", seconds );
      oled.setCursor( 42, 1 );
      oled.println( (String) hours + "h " + min_char + "m " + sec_char + "s" );
      if( total_counts )
      {
        current_time -= total_deadtime;
        float_time = .001 * current_time;
        count_average = (float) total_counts / float_time;
        count_std = sqrt( (float) total_counts ) / float_time;
        if( count_std < .001 )
          decimals = 5;
        else
        if( count_std < .01 )
          decimals = 4;
        else
        if( count_std < .1 )
          decimals = 3;
        else
        if( count_std < 1. )
          decimals = 2;
        else
          decimals = 1;
        dtostrf( count_average, 1, decimals, average_char );
        dtostrf( count_std, 1, decimals, stddev_char );
        oled.print( (String) average_char );
        oled.print( F("+/-") );
        oled.print( (String) stddev_char );
        oled.print( F(" cps   ") );
      }
      screen_time = start_deadtime + TIMER_INTERVAL;
      stop_deadtime = millis();
      if( stop_deadtime < start_deadtime )
      {
        stop_deadtime += 2147483648;
        start_deadtime -= 2147483648;
      }
      total_deadtime += stop_deadtime - start_deadtime;
    }
  }
}

boolean get_detector_name(char* det_name) 
{
  byte ch;
  int bytesRead = 0;
  do {
    ch = EEPROM.read( bytesRead );
    det_name[bytesRead] = ch;
    bytesRead++;
  } while( ( ch != 0x00 ) && ( bytesRead < DETECTOR_NAME_SIZE ) );
  if( ( ch != 0x00 ) && ( bytesRead > 0 ) )
    det_name[bytesRead - 1] = 0;
  return true;
}

byte next_filename()
{
  if( filename[7] < '9' )
    filename[7]++;
  else
  {
    filename[7] = '0';
    if( filename[6] < '9' )
      filename[6]++;
    else
    {
      filename[6] = '0';
      if( filename[5] < '9' )
        filename[5]++;
      else
      {
        filename[5] = '0';
        return 1;
      }
    }
  }
  return 0;
}

float get_sipm_voltage(float adc)
{
  float x = cal[0] * adc;
  x += cal[1];
  x *= adc;
  x += cal[2];
  x *= adc;
  x += cal[3];
  x *= adc;
  x += cal[4];
  x *= adc;
  x += cal[5];
  x *= adc;
  x += cal[6];
  return x;
}
